__version__ = "3.4.1"

DEFAULTS = {
    "dump_only_pk": True,
    "string_keys": True,
    "id_keys": False,
}

from .fields import ForeignKey, MSTimestamp, Related, Timestamp
from .schema import ModelSchema


def setup(**options):
    """Setup marshmallow-peewee defaults."""
    DEFAULTS.update(options)


__all__ = "ModelSchema", "Timestamp", "MSTimestamp", "Related", "ForeignKey", "setup"



import datetime as dt
import typing as t

import peewee as pw
from marshmallow import Schema, fields


class Timestamp(fields.Field):

    default_error_messages = {"invalid": "Not a valid timestamp."}

    def _serialize(
        self, value: dt.datetime, attr: t.Optional[str], obj: pw.Model, **kwargs
    ) -> t.Optional[int]:
        """Serialize given datetime to timestamp."""
        if value is None:
            return None

        return int(value.timestamp())

    def _deserialize(
        self,
        value: t.Union[str, int, float],
        attr: t.Optional[str],
        data: t.Optional[t.Mapping[str, t.Any]],
        **kwargs,
    ) -> dt.datetime:
        if not value:  # Falsy values, e.g. '', None, [] are not valid
            raise self.make_error("invalid")

        try:
            return dt.datetime.utcfromtimestamp(float(value))
        except ValueError:
            raise self.make_error("invalid")


class MSTimestamp(Timestamp):
    def _serialize(
        self, value: dt.datetime, attr: t.Optional[str], obj: pw.Model, *_, **__
    ) -> t.Optional[int]:
        """Serialize given datetime to timestamp."""
        val = super(MSTimestamp, self)._serialize(value, attr, obj)
        if val is None:
            return None

        return val * 1000

    def _deserialize(
        self,
        value: t.Union[str, int, float],
        attr: t.Optional[str],
        data: t.Optional[t.Mapping[str, t.Any]],
        **_,
    ) -> dt.datetime:
        if value:
            value = int(value) / 1e3

        return super(MSTimestamp, self)._deserialize(value, attr, data)


class Related(fields.Nested):
    def __init__(self, nested: Schema = None, meta: t.Dict = None, **kwargs):  # type: ignore
        self.field = None
        self.meta = meta or {}
        super(Related, self).__init__(nested, **kwargs)  # type: ignore

    def init_model(self, model: pw.Model, name: str):
        from .schema import ModelSchema

        field = model._meta.fields.get(name)  # type: ignore
        if not field:
            field = getattr(model, name).field
            self.many = True
            rel_model = field.model
        else:
            rel_model = field.rel_model

        self.field = field
        self.attribute = self.attribute or name
        self.meta["model"] = rel_model
        meta = type("Meta", (), self.meta)
        self.nested = type("Schema", (ModelSchema,), {"Meta": meta})

    def _deserialize(self, value, attr, data, **_):
        if self.field is None:
            raise RuntimeError("Init model first.")

        if not isinstance(value, dict):
            return self.field.rel_field.python_value(value)

        return super(Related, self)._deserialize(value, attr, data)


class ForeignKey(fields.Raw):
    string_keys = False

    def _bind_to_schema(self, field_name, schema):
        self.string_keys = schema.opts.string_keys
        super()._bind_to_schema(field_name, schema)

    def get_value(self, obj: pw.Model, attr: str, **_) -> t.Any:  # type: ignore
        """Return the value for a given key from an object."""
        value = obj.__data__.get(self.attribute)
        if value is not None and self.string_keys:
            value = str(value)
        return value



from __future__ import annotations

import typing as t
from collections import OrderedDict

import peewee as pw
from marshmallow import ValidationError, fields
from marshmallow import validate as ma_validate

from .fields import ForeignKey


class ModelConverter:
    """Convert Peewee model to Marshmallow schema."""

    TYPE_MAPPING = [
        (pw.IntegerField, fields.Integer),
        (pw.DateTimeField, fields.DateTime),
        (pw.DateField, fields.Date),
        (pw.TextField, fields.String),
        (pw.ForeignKeyField, ForeignKey),
        (pw.FloatField, fields.Float),
        (pw.DecimalField, fields.Decimal),
        (pw.TimeField, fields.Time),
        (pw.BigIntegerField, fields.Integer),
        (pw.SmallIntegerField, fields.Integer),
        (pw.DoubleField, fields.Float),
        (pw.FixedCharField, fields.String),
        (pw.UUIDField, fields.UUID),
        # Overrided
        #  (pw.AutoField, fields.String),
        #  (pw.BigAutoField, fields.String),
        #  (pw.CharField, fields.String),
        #  (pw.BooleanField, fields.Boolean),
    ]

    def __init__(self, opts: SchemaOpts):
        self.opts = opts

    def fields_for_model(self, model: pw.Model) -> OrderedDict:
        result = OrderedDict()
        for field in model._meta.sorted_fields:
            name = field.name
            if self.opts.id_keys and isinstance(field, pw.ForeignKeyField):
                name = field.column_name

            ma_field = self.convert_field(field)
            if ma_field:
                result[name] = ma_field

        return result

    def convert_field(self, field: pw.Field) -> fields.Field:
        params = {
            "allow_none": field.null,
            "attribute": field.name,
            "required": not field.null and field.default is None,
            "validate": [convert_value_validate(field.db_value)],
        }

        if field.null:
            params["load_default"] = None

        if field.default is not None:
            params["load_default"] = field.default

        if field.choices is not None:
            choices = []
            labels = []
            for c in field.choices:
                choices.append(c[0])
                labels.append(c[1])
            params["validate"].append(ma_validate.OneOf(choices, labels))

        if field.help_text:
            params["metadata"] = {"description": field.help_text}

        # use first "known" field class from field class mro
        # so that extended field classes get converted correctly
        method = None
        for klass in field.__class__.mro():
            method = getattr(self, "convert_" + klass.__name__, None)
            if method:
                break

        method = method or self.convert_default
        return method(field, **params)

    def convert_default(self, field: pw.Field, **params) -> fields.Field:
        """Return raw field."""
        for klass, ma_field in self.TYPE_MAPPING:
            if isinstance(field, klass):
                return ma_field(**params)
        return fields.Raw(**params)

    def convert_AutoField(
        self, field: pw.Field, required: bool = False, **params
    ) -> fields.Field:
        ftype = fields.String if self.opts.string_keys else fields.Integer
        return ftype(dump_only=self.opts.dump_only_pk, required=False, **params)

    convert_BigAutoField = convert_AutoField

    def convert_CharField(
        self, field: pw.CharField, validate: t.List = None, **params
    ) -> fields.Field:
        if validate is None:
            validate = []

        validate += [ma_validate.Length(max=field.max_length)]
        return fields.String(validate=validate, **params)

    def convert_BooleanField(
        self, field: pw.Field, validate: t.List = None, **params
    ) -> fields.Field:
        return fields.Boolean(**params)


def convert_value_validate(converter: t.Callable) -> t.Callable:
    def validator(value):
        try:
            converter(value)
        except Exception as e:
            raise ValidationError(str(e))

    return validator


if t.TYPE_CHECKING:
    from .schema import SchemaOpts
